' *******************************************************************************************************
' *                                                                                                     *
' *     AiChip_USB_Driver_001.spin                                                                      *
' *                                                                                                     *
' *******************************************************************************************************
' .-----------------------------------------------------------------------------------------------------.
' |                                                                                                     |
' |     original : USB_LowLevel_80_008.spin                                                             |
' |                                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

' Copyright Brad Campbell, 2007 - All rights reserved
' This code is not USB compliant. Contact the author for any kind of commercial use.

' 3 IO pins used. Enable, D+, D-
' Current configuration.

' 0 - D-
' 1 - D+
' 2 - Enable

' Tie 2 to D- (at the cable, not the Propeller) with a 1.5k resistor and put a pair of 68Ohm resistors
' in series with D+ and D- to the host

' I've used 45-100Ohm for the series resistor, but 1.2-1.5k is about right for the enable pullup.
' 3.3k is _too_ high (as is 2.2k). USB spec says between 900 Ohm and 1.5k.

' While the specification dictates we must turn around in 6.5 or 7.5 bit times, it also says a host must
' wait at least 16 and not more than 18 bit times for a response.

' 5 hubs + cable propagation delay is 760ns which equates to 1.14 bit times, leaving us 14.8 bit times
' to effectively turn around in. This is about 196 instructions. 

' A DATA1 takes up to 19.6 bit times to process (261 instruction cycles). We now send the Ack immediately we've processed
' the packet and write to the buffer (3 bit times) during the EOP of the ack. Ths brings us within 16 cycles. Given we
' ignore the inbound EOP giving us 18-20 bit times to turn around in, we now fall within the spec.

' For a full buffer DATA1 it takes a maximum of 231 instruction cycles to start the K of the ack.
' This is 17.3 bit times from start of inbound EOP. We have 18-20 available to us.

' While most other software USB SIO engines fall within the spec for timing, they ignore CRC/EOP/STUFF/Babble errors
' to do so. We fall outside the 7.5 bit time return (however within the 16 bit time maximum limit) and check for all
' known error states.

' Endpoint 0 has a common TX/RX toggle bit. All other endpoints have separate TX/RX toggle bits

' Rev 6 - Windows runs IN packets all the time. An unplug while recieving an IN packet would cause the RX routine
' to go haywire and not detect an EOP, causing it to think it was recieving lots of bytes and scribling all over
' RAM. We now detect the bitstuff error that this triggers and abort reception. Given we abort on a bitstuff error
' in any case, this solves the issue
' Released 16/09/2007

' Rev 7 - More accurate start bit positioning
'       - Pass pins down through start()
'       - Interleave inbound buffer writing in with transmitting the EOP and stop J to try and speed up
'       - .. getting an ack out and get closer to spec timing.
'       - Filter out ACK packets that were not intended for us.
'       - Filter out DATA packets that were not intended for us. This makes it co-exist on the same bus as my APC UPS
'       - Released 20/09/2007

' Rev 8 - Rationalise inbound packet DATA and ACK checking to save code space
'       - Some more code consolidation and hacks to save space
'       - Added Babble detection/protection
'       - Improved reset detection/handling
'       - Removed bitcnt rounding
'       - Pulled rx bit position back to cope with VIA chipsets
'       - Doc change, Enable must be a 1.5k - 3.3k is too high and causing VIA hardware on Windows to disconnect randomly!
'       - Properly ignore errors and ensure we don't accept data packets without a valid token 1st
'       - Don't accept an address packet if we are already addressed (Must be reset first)
'       - Option to pace transmit packets to work around bugs in old USB silicon coupled with Windows drivers
'       - Added parameter in Start() for packet pacing (VIA bugfix)
'       - 14 longs free
'       - Released 30/09/2007

' TODO  - Further improve/rationalise reset detection

' *******************************************************************************************************
' *                                                                                                     *
' *                                                                                                     *
' *                                                                                                     *
' *******************************************************************************************************

PUB Start(crctable, buffers, Pdminus, Pdplus, Penable, pace) : OK | i
  Pcrcmask := crctable
  epb := buffers
  ' IO Pin configuration
  ' D+
  K := 1 << Pdplus
  ' D-
  J := 1 << Pdminus     
  ' Enable 3.3k to D-
  Enable := 1 << Penable  
  If pace <> 0
    pace := 1 << 2
  OK := cognew(@USB_STACK, pace)


DAT

' *******************************************************************************************************
' *                                                                                                     *
' *                                                                                                     *
' *                                                                                                     *
' *******************************************************************************************************

USB_STACK     org
              ' If par <> 0 then we need to pace our transmission
              test      allones, par wz 
        IF_NZ movs      wpace, #pid_in_paced
              mov       bmask, K
              or        bmask, J
              call      #crc5_address
              wrlong    allones, epb            ' Semaphore to wait for app thread to be running
:sloop        rdlong    da, epb  wz
        IF_NZ jmp       #:sloop                 ' Loop until it is
              or        outa, enable
              or        dira, enable            ' Tell the host we are alive!
              jmp       #reset_here

' --------------------------------------------
' Receive low level routines
' --------------------------------------------


rxreload      add       bufptr, #1              ' 12 post inc pointer to allow next mov to pipeline
rxindr        mov       0-0, da                 ' 13 store contents of buffer
              mov       bitcnt, #32             ' 14 reload bit counter
              
rx            test      K, ina  wz              ' 1  - 13 bits
              muxz      db, emask wc            ' 2  - top 2 bits contain previous and new values..
              shl       db, #1                  ' 3  - parity check 00 or 11 is odd and 01 or 10 is even
              test      bmask, ina wz           ' 4  - Check for EOP
        IF_Z  jmp       #wfe                    ' 5
              rcr       da, #1                  ' 6  - rotate carry from xor into top bit of acc
              rcl       dc, #1                  ' 7  - Bottom 7 bits are all 0 we need to unstuff
              and       dc, #$3F wz             ' 8  - test is inverted as input bits are inverted
        IF_Z  call      #rxunstuff              ' 9  - we need to unstuff next bit
              sub       bitcnt, #1 wz           ' 10 
        IF_Z  jmp       #rxreload               ' 11
              movd      rxindr, bufptr          ' 12 - This saves a cycle in rxreload and does not hurt to keep repeating it
              nop                               ' 13
                         
              test      K, ina  wz              ' 1  - 13 bits
              muxz      db, emask wc            ' 2  - top 2 bits contain previous and new values..
              shl       db, #1                  ' 3  - parity check 00 or 11 is odd and 01 or 10 is even
              test      bmask, ina wz           ' 4  - Check for EOP
        IF_Z  jmp       #wfe                    ' 5
              rcr       da, #1                  ' 6  - rotate carry from xor into top bit of acc
              rcl       dc, #1                  ' 7  - Bottom 7 bits are all 0 we need to unstuff
              and       dc, #$3F wz             ' 8  - test is inverted as input bits are inverted
        IF_Z  call      #rxunstuff              ' 9  - we need to unstuff next bit
              sub       bitcnt, #1 wz           ' 10 
        IF_Z  jmp       #rxreload               ' 11
              movd      rxindr, bufptr          ' 12
              nop                               ' 13
                            
              test      K, ina  wz              ' 1  - 14 bits
              muxz      db, emask wc            ' 2  - top 2 bits contain previous and new values..
              shl       db, #1                  ' 3  - parity check 00 or 11 is odd and 01 or 10 is even
              test      bmask, ina wz           ' 4  - Check for EOP
        IF_Z  jmp       #wfe                    ' 5
              rcr       da, #1                  ' 6  - rotate carry from xor into top bit of acc
              rcl       dc, #1                  ' 7  - Bottom 7 bits are all 0 we need to unstuff
              and       dc, #$3F wz             ' 8  - test is inverted as input bits are inverted
        IF_Z  call      #rxunstuff              ' 9  - we need to unstuff next bit
              sub       bitcnt, #1 wz           ' 10 
        IF_Z  jmp       #rxreload               ' 11
              movd      rxindr, bufptr          ' 12
              cmp       bufptr, #addr wz        ' 13 - Was a nop ' This is Babble detection. 
        IF_NZ jmp       #rx                     ' 14
              ' If we get to here we have babble/dribble and would be overwriting memory!!
              jmp       #error
                      
              ' The following 2 instructions rely on the fact they take 8 cycles instead of 4 knowing
              ' full well they will _never_ be zero and thus give us a 4 nop delay for 2 instructions
rxunstuff     tjz       allones, #rxunstuff     ' 10 | 11
              tjz       allones, #rxunstuff     ' 12 | 13
              test      K, ina  wz              ' 1
              muxz      db, emask wc            ' 2
              shl       db, #1                  ' 3
              rcl       dc, #1                  ' 4
              test      bmask, ina wz           ' 5
        IF_Z  jmp       #wfe                    ' 6
              test      dc, #$7F wz             ' 7
        IF_Z  jmp       #error                  ' 8     ' Bitstuffing error
rxunstuff_ret ret                               ' 9


error
              andn      state, flag_waitp       ' If we are in error then we won't be listening
              ' Fall through to #wfs            ' for the next packet.
                                                
' Wait For Start
' Sync preamble is KJKJKJKK - We just look for 2xK following a J and we are there
' Use waitpeq to start off cycle accurate as possible
wfs           mov       reset, #17              ' 17 * 3 instructions = 3.8 bit times              
              mov       bitcnt, #32             
              mov       dc, #$FF                ' - Stuff register is inverted
              mov       bufptr, #buf0           ' Init indirect buffer pointer
              mov       db, #0                  ' Zero xor register
              mov       flags, #0               ' Set up status flags
              mov       addr, #0                ' Hub address used in pid_in
:wfs0         mov       da, ina
              and       da, bmask  wz
        IF_Z  jmp       #wait_reset             ' SE0
              cmp       da, K      wz
        IF_NZ jmp       #:wfs0                  ' K
:wfs1         waitpeq   J, bmask
:wfs2         waitpeq   K, bmask
              mov       dd, cnt
              add       dd, #10                 ' Delay half bit time - calcs | This is experimentally derived!
                                                ' The calculated value works well on most chipsets. VIA chipsets
                                                ' require this lower value to work reliably and it does not impact
                                                ' any other chipsets I've tested on.
              waitcnt   dd, #53
              mov       da, ina
              and       da, bmask
              cmp       da, K wz
        IF_NZ jmp       #:wfs1                  ' K
              waitcnt   dd, #53
              mov       da, ina
              and       da, bmask
              cmp       da, K wz
        IF_NZ jmp       #:wfs1                  ' K
              waitcnt   dd, #54
              jmp       #rx


wait_reset    sub       reset, #1  wc           ' If SE0 goes on longer than 50 instruction cycles then it's a reset and we process it
              test      bmask, ina  wz
  IF_Z_AND_NC jmp       #wait_reset             ' Still in SE0 and "reset" > 0
 IF_NZ_AND_NC jmp       #wfs                    ' Not in SE0 and "reset" > 0    ' Was probably a periodic sync
reset_here    wrlong    flag_reset, epb         ' Tell the upper thread to reset
              mov       address, #0             ' Device Address
              mov       eps, #$FF               ' Endpoint toggle bits
              mov       state, #0               ' Device status flags
              mov       paddr, #0               ' Pending address 
              call      #crc5_address           ' Reset crc buffer to address 0
              waitpeq   J, bmask                ' Wait for reset state to end
              jmp       #wfs
              

' --------------------------------------------
' Transmit low level routines
' Transmit speed is fast by 0.625% (spec says we can be up to 1.5% so it's all good)
' Most of the transmit loop is instruction critical and it *will* lock up if you get it wrong
' If you are lucky it will lock up on a J, but if it locks up on a K or SOP it *will* take out the
' USB port and require a restart. Testing through a HUB is a good idea as you can cycle the HUB.
' If you are using Windows, it's random but the odds are you will have to reboot anyway.
' --------------------------------------------


' Bitstuff the next bit
txs           waitcnt   delay, #53
              xor       outa, bmask
              shl       db, #1
              or        db, #1
txs_ret       ret


' Move onto the next byte but invert it first (we transmit inverted to make stuffing easy)
txreload      mov       bitcnt, #8
txreload_ind  mov       da, 0-0
              sub       bytecnt, #1  wz
        IF_Z  jmp       #tx_eop                 ' Put this here rather than the end to give us
              xor       da, #$FF                ' A spare cycle or two ensuring the EOP waitcnt
              add       txreload_ind, #1        ' does not hang

tx1           waitcnt   delay, #53
stx1          rcr       da, #1 wc
        IF_C  xor       outa, bmask
              rcl       db, #1
              and       db, #$3F wz
        IF_Z  call      #txs
tx1e          sub       bitcnt, #1 wz

              test      da, #1 wz               ' Placing this here gives us a little extra
              shl       db, #1                  ' room in the reload routine to allow xor of
              muxnz     db, #1                  ' data and proper indirect loading
              and       db, #$3F wz             ' Transmitting pairs of bits ping/pong buys us this
              sub       bitcnt, #1
              
tx2           waitcnt   delay, #53
              rcr       da, #1 wc
        IF_C  xor       outa, bmask
              tjnz      db, #tx2r    
              call      #txs
tx2r          tjz       bitcnt, #txreload
              jmp       #tx1


tx_eop        waitcnt   delay, #106             ' This delay is *tight*
              andn      outa, bmask             ' Start the EOP
              test      addr, allones wz
        IF_NZ jmp       #save_data              ' If we have data to write to the upper level, we best do it
              waitcnt   delay, #53
              or        outa, J                 ' Force a J
tx_ret        waitcnt   delay, #53              ' Delay one bit time... and
              andn      dira, bmask             ' Switch off the drivers
              andn      outa, bmask
              jmp       #wfs
              
do_tx         mov       bitcnt, #8 
              mov       da, #$7F                ' Sync byte (inverted as we never xor that one)
              movs      txreload_ind, #b_tkn    ' Set up indirect reload register
              add       bytecnt, #1             ' Add sync byte to count
              mov       db, #$FF                ' Stuff register is inverted
              andn      outa, bmask             ' Clear D+ / D- Pins
              or        outa, J                 ' Set pins up as a J before we enable the drivers
              or        dira, bmask             ' enabling the drivers
              mov       delay, cnt
              add       delay, #57              ' Proper time for us to get settled
              jmp       #stx1                   ' .... and away we go

' --------------------------------------------
' Protocol stack routines
' --------------------------------------------

' Responses to setup queries in the STATUS phase must go out as DATA1
send_empty_data1
              mov       b_1, #0
              mov       b_2, #0
              mov       b_tkn, #$4B
              mov       bytecnt, #3
              or        state, flag_waitp       ' We are going to want an ack for this
              jmp       #do_tx
              
send_ack      mov       b_tkn, #$D2             ' PID ACK
send_any      mov       bytecnt, #1
              jmp       #do_tx

send_nak      mov       b_tkn, #$5A             ' PID NAK
              jmp       #send_any               ' This hack courtesy of insufficient code space !!

pid_setup     or        state, flag_setup
              or        eps, #$11               ' Set up toggle bits
pid_setout    or        state, flag_waitp       ' Next data packet is for us
              jmp       #wfs                     

pid_out       andn      state, flag_setup
              jmp       #pid_setout             ' Gross space saving hack

pid_in_paced
              ' Ugly way of slowing down transmission. Only send on every other IN token
              mov       da, #$100               ' Start at bit 8 (bit's 0-7 are endpoint toggle bits)
              shl       da, epoint
              xor       eps, da
              test      eps, da wz
        IF_Z  jmp       #send_nak
pid_in        mov       da, epoint
              shl       da, #4                  ' Multiply by 8
              add       da, epb                 ' Add eps offset
              add       da, #64                 ' Output endpoints
              rdlong    db, da  wz              ' Endpoint out byte counter
        IF_Z  jmp       #send_nak               ' Nothing to send
              mov       dc, db
              andn      dc, #$FF                ' Clear the bottom 8 bits (byte counter)
              or        state, dc               ' Set any flags passed down from the application stack
              test      db, flag_empty1 wz      ' Are we in the STATUS phase of a setup query?
        IF_NZ jmp       #send_empty_data1       ' ...  Send an empty DATA1 packet
              test      db, flag_empty  wz      ' Send an empty packet with the next logical DATA0/1 sequence
        IF_NZ jmp       #:pid_send_empty        ' Token (BULK/INT out where last data was packet aligned)
              mov       bytecnt, db
              and       bytecnt, #$0F           ' Clamp max in case app level goes haywire
              add       da, #4                  ' Unpack the bytes from here down (inline for speed)
              rdlong    db, da                  ' Endpoint out bytes [0-3]
              mov       b_1, db
              shr       db, #8
              mov       b_2, db
              shr       db, #8
              mov       b_3, db
              shr       db, #8
              mov       b_4, db
              add       da, #4
              rdlong    db, da                  ' Endpoint out bytes [4-7]
              mov       b_5, db
              shr       db, #8
              mov       b_6, db
              shr       db, #8
              mov       b_7, db
              shr       db, #8
              mov       b_8, db
              add       da, #4
              rdlong    db, da                  ' Endpoint out bytes [8-9]
              mov       b_9, db
              shr       db, #8
              mov       b_10, db
              jmp       #:pid_next
:pid_send_empty
              mov       b_1, #0                 ' This is a cheat, but the CRC on an empty packet is $00 $00
              mov       b_2, #0
              mov       bytecnt, #2
:pid_next
              mov       da, #1
              shl       da, epoint
              test      epoint, allones wz
        IF_NZ shl       da, #4
              test      eps, da wz              ' Make sure data toggle is correct for token
        IF_NZ mov       b_tkn, #$C3
        IF_Z  mov       b_tkn, #$4B
              add       bytecnt, #1             ' Allow for token
              or        state, flag_waitp       ' We will want an ACK after this one
              jmp       #do_tx              

pid_data      ' 192 to here (DATA1)
              andn      state, flag_waitp
              test      state, flag_setup wz
        IF_Z  jmp       #:data_1                ' Is this a setup packet??
              or        flags, flag_setup       ' Tell upper layer this is a setup request
              test      b_1, #$20 wz            ' Class packet?
        IF_NZ or        flags, flag_class       ' Tell upper layer this is a class request
              mov       da, b_2                 ' Get bREQUEST
              and       da, #$FF

              cmp       da, #5     wz           ' Is this a SET_ADDRESS request?
        IF_NZ jmp       #:data_1
              test      Address, allones wz
        IF_NZ jmp       #error                  ' We are already addressed. Not again until we are reset
              mov       da, b_3                 ' Get new address
              and       da, #$7F                ' Addresses are 7 bits
              mov       paddr, da               ' Make it pending until application layer has completed transaction
:data_1
              mov       da, epoint
              shl       da, #4                  ' Multiply by 4
              add       da, epb                 ' Add endpoint buffer offset
              rdlong    db, da  wz              ' Endpoint byte counter
        IF_NZ jmp       #send_nak               ' Other thread is not finished with this buffer
              mov       addr, da                ' Save data for use in the ACK's EOP
              mov       inbytecnt, bytecnt      ' This saves the bytecnt being overwritten while we send the ACK
              jmp       #send_ack               ' Get an ack out as quickly as possible ' 218 instructions to here
              
              ' Shuffle to convert buffers (we need only 1st 8 bytes as we don't want CRC anyway)
              ' IN  X987 6543 210X         (Byte X - next to byte 0 - is the token. Not required)
              ' OUT 7654 3210

save_data     ' This is processed while we are sending the EOP for the ack packet

              mov       db, buf0
              shr       db, #8
              mov       dc, buf1
              shl       dc, #24
              or        db, dc                  ' db is 1st long
              mov       dd, buf1
              shr       dd, #8
              mov       dc, buf2
              shl       dc, #24
              or        dd, dc                  ' dd is 2nd long

              ' Write buffer for notice of application thread
              ' Always write byte count / flags last to ensure we don't try to read while
              ' the buffer is half written
              
              add       addr, #4                ' Endpoint bytes [0-3] 
              wrlong    db, addr                ' Write 1st long
              add       addr, #4                ' Endpoint bytes [4-7]
              sub       inbytecnt, #3           ' Minus token and CRC
              wrlong    dd, addr                ' Write 2nd long
              sub       addr, #8
              or        inbytecnt, flags

              waitcnt   delay, #53              ' Wait for 106 cycles of the EOP ' We now have 53 cycles to get finished
              or        outa, J                 ' Force a J

              wrlong    inbytecnt, addr         ' Number of bytes in
              andn      state, flag_setup       ' Clear the setup flag if it was set
              mov       da, #1
              shl       da, epoint
              xor       eps, da                 ' Toggle recieve toggle bit
              mov       addr, #0                ' Make sure we are right for the next transmission and don't end up back here
              jmp       #tx_ret

              
pid_dataB     ' 26 to here for a DATA1
              ' Unpack the remainder of the inbound data
              xor       buf1, allones
              xor       buf2, allones
              mov       da, buf1
              mov       b_4, da
              shr       da, #8
              mov       b_5, da
              shr       da, #8
              mov       b_6, da
              shr       da, #8
              mov       b_7, da
              mov       da, buf2
              mov       b_8, da
              shr       da, #8
              mov       b_9, da
              shr       da, #8
              mov       b_10, da                ' 42 To here
              call      #crc16check             ' 185 to here
              test      flags, flag_crc  wz
        IF_NZ jmp       #error                  ' Packet was bad
              mov       da, #1
              shl       da, epoint
              mov       db, #$4d
              cmp       db, b_tkn wc            ' Carry set if eq $C3 - DATA0
              test      eps, da wz
 IF_NZ_AND_NC jmp       #send_ack               ' DATA1 & We already have this data packet
   IF_Z_AND_C jmp       #send_ack               ' DATA0 & We already have this data packet
              jmp       #pid_data


pid_ack
              mov       da, #1
              shl       da, epoint
              test      epoint, allones wz
        IF_NZ shl       da, #4
              xor       eps, da                 ' Toggle endpoint state
              mov       da, epoint
              shl       da, #4                  ' Multiply by 8
              add       da, epb                 ' Add eps offset
              add       da, #64                 ' Output endpoints
              wrlong    zero, da                ' Tell other thread we are ready for more data
              andn      state, flag_waitp
              test      state, flag_address wz  ' on this endpoint
        IF_Z  jmp       #wfs
              mov       address, paddr          ' Set the new address
              andn      state, flag_address     ' Clear the flag baby!
              call      #crc5_address
              jmp       #wfs


wfe           ' Wait For End
              ' We ignore the test for EOP and J, this buys us an extra 2 bit times to turnaround
              movd      :wind, bufptr
              ' Shift final word to sit properly in lsb order
              shr       da, bitcnt
:wind         mov       0-0, da
              ' Fall through to process_inbound

' This is where we check the inbound packet for validity
process_inbound        
              xor       buf0, allones
              ' Unpack just as much as we may need if it's not a DATA packet
              ' We do all bytes of the 1st long to save a reload later
              mov       da, buf0
              mov       b_tkn, da
              shr       da, #8
              mov       b_1, da
              shr       da, #8
              mov       b_2, da
              shr       da, #8
              mov       b_3, da

              ' This next bit is essential as the hub code uses it to tell how many longs to pull
              ' and display as this inbound message
              ' This calculates the number of bytes in the message

              mov       bytecnt, bufptr
              sub       bytecnt, #buf0
              shl       bytecnt, #2
              mov       da, #$20
              sub       da, bitcnt
              shr       da, #3
              or        bytecnt, da
              and       b_tkn, #$FF
              ' 20 Instructions from start of EOP to here
              
              ' data in and data out need address and endpoint of last token packet
              ' Hopefully set last rx packet
              test      state, flag_waitp  wz
        IF_Z  jmp       #:next                  ' We are not waiting for a data or an Ack

              cmp       b_tkn, #$c3 wz          ' Data0
        IF_Z  jmp       #pid_dataB
              cmp       b_tkn, #$4b wz          ' Data1
        IF_Z  jmp       #pid_dataB
              cmp       b_tkn, #$D2 wz          ' Ack
        IF_Z  jmp       #pid_ack

:next         mov       da, b_1
              and       da, #$7F
              cmp       da, address wz
        IF_NZ jmp       #error                  ' Not addressed to us!
              call      #crc5check
              test      flags, flag_crc wz      ' Token failed CRC check
        IF_NZ jmp       #error
        
              mov       epoint, b_1
              shr       epoint, #7
              and       epoint, #$0F
              test      epoint, #$0C wz
        IF_NZ jmp       #error                  ' Not one of our endpoints!

              cmp       b_tkn, #$E1 wz          ' OUT
        IF_Z  jmp       #pid_out
              cmp       b_tkn, #$69 wz          ' IN
wpace   IF_Z  jmp       #pid_in
              cmp       b_tkn, #$2D wz          ' Setup
        IF_Z  jmp       #pid_setup
              jmp       #error                  ' Obviously not a valid token

' Calculate crc of packet using pre-calculated address crc and comparing the last 5 bits
' Now fast enough to be used in line.
crc5check     ' 43 Instruction cycles
              mov       crc, crc5addr
              mov       da, #4
              mov       dc, b_1
              shr       dc, #7
              call      #crc5bit                
              mov       da, crc
              shl       crc, #3
              xor       crc, allones
              xor       crc, b_2
              and       crc, #$F8  wz
        IF_NZ or        flags, flag_crc
crc5check_ret ret              
              

' Precaculate the address into the CRC register when we change address to save 7 loops
' of the crc routine
crc5_address  mov       crc, #$1F
              mov       dc, paddr
              mov       da, #7
              call      #crc5bit
              mov       crc5addr, crc
crc5_address_ret
              ret

' This is horridly slow.
crc5bit      ' dc has the byte to add to the crc | da has the number of bits to process
              and       dc, #$FF
crc5bitloop   mov       db, dc
              xor       db, crc
              shr       dc,  #1
              shr       crc, #1
              shr       db,  #1 wc
        if_c  xor       crc, #$14
              djnz      da, #crc5bitloop
crc5bit_ret ret

' Perform a crc check on recieved data packets
' Mangles da, db, dc
' Returns crc          

crcmask       long      $0000FFFF

crc16check    ' 142 Instruction cycles for a full buffer
              mov       crc, crcmask
              movs      :crc16indr, #b_1              
              mov       da, bytecnt
              sub       da, #3
              test      da, allones  wz
        IF_Z  jmp       #crc16check_ret         ' This dies horribly if we pass it a no-data packet
:crc16indr    mov       dc, 0-0
              add       :crc16indr, #1
              xor       dc, crc                 
              and       dc, #$FF                
              shl       dc, #1                  
              add       dc, Pcrcmask            
              rdword    db, dc                  
              shr       crc, #8                 
              xor       crc, db                 
              djnz      da, #:crc16indr
' This extra complexity saves us 2 loops (up to 30 instruction cycles and costs us 12 extra)
' Net gain of 18 instructions or 1.5 bit times
              xor       crc, allones
              and       crc, crcmask
              mov       :crc16indr1, :crc16indr
              add       :crc16indr, #1
:crc16indr1   nop                               ' Indirect move
              and       dc, #$FF                ' 2
              mov       :crc16indr2, :crc16indr               
              mov       da, dc
:crc16indr2   nop                               ' Indirect move
              and       dc, #$FF                ' 2
              shl       dc, #8
              or        dc, da
              cmp       crc, dc wz
        IF_NZ or        flags, flag_crc
crc16check_ret ret

reset         long      0
allones       long      $FFFFFFFF
zero          long      $0

' These 5 are loaded in spin prior to starting the cog
Pcrcmask      long      0                       ' Pointer to CRC table in hub
K             long      0                       ' D+ line
J             long      0                       ' D- line
enable        long      0                       ' Enable enumeration pullup
epb           long      0                       ' Endpoint communications buffer [32 longs]                       

bmask         long      0                       ' Bitmask of K & J together - Calculated at cog start
emask         long      $40000000               ' MUX mask for RX inbound xor register

'flag_comms   long      1 << 21
flag_reset    long      1 << 22
flag_rtoggle long      1 << 23
flag_empty1   long      1 << 24                 ' Send an empty DATA1 packet
'flag_debug   long      1 << 25                 ' Debugging used in the app level
flag_class    long      1 << 26                 ' We recieved a CLASS request
flag_address  long      1 << 27                 ' Inbound setup packet is an address request
flag_empty    long      1 << 28                 ' Send next logical DATAX token in an empty packet
flag_setup    long      1 << 29                 ' Inbound packet is a setup request
flag_waitp    long      1 << 30                 ' Are we waiting for a data or ack packet?
flag_crc      long      1 << 31                 ' CRC error detected

eps           res       1                       ' endpoint status
address       res       1                       ' device address
epoint        res       1                       ' inbound endpoint
state         res       1                       ' State machine flags
paddr         res       1                       ' Pending address to be set at end of STATUS cycle

bitcnt        res       1

' Unpacked data buffer
b_tkn         res       1
b_1           res       1
b_2           res       1
b_3           res       1
b_4           res       1
b_5           res       1
b_6           res       1
b_7           res       1
b_8           res       1
b_9           res       1
b_10          res       1

' Packed data buffer
buf0          res       1                       ' TX/RX Buffer
buf1          res       1
buf2          res       1

' Used to hold hub address when we write packet while sending EOP    
' This is also used as a guard to detect babble. If we hit this addrerss while we are in rx then things have gone very wrong
' and we abort the recieve before we trample all over RAM
addr          res       1                       

' Indirect buffer pointer
bufptr        res       1                   

' Misc temporary registers
da            res       1
db            res       1
dc            res       1
dd            res       1

' Recieve status flags
flags         res       1                       

' TX / RX bytes in or bytes to send
bytecnt       res       1

' Cumulative crc register
crc           res       1

' Pre-calculated crc load based on device address
crc5addr      res       1

' Separate delay variable to save code when transitioning from sending EOP to writing buffer
delay         res       1                       

' Separate counter for number of inbound bytes to save being overwritten while sending ACK

inbytecnt     res       1                       

              fit       $1F0

' *******************************************************************************************************
' *                                                                                                     *
' *     End of AiChip_USB_Driver_001.spin                                                               *
' *                                                                                                     *
' *******************************************************************************************************
